"""Logic for interacting with sphinx-build."""

from __future__ import annotations

import contextlib
import subprocess
import sys
import traceback
from collections.abc import Sequence
from pathlib import Path

import sphinx

from sphinx_autobuild.utils import show_command, show_message


class Builder:
    def __init__(self, sphinx_args, *, url_host, pre_build_commands):
        self.sphinx_args = sphinx_args
        self.pre_build_commands = pre_build_commands
        self.uri = f"http://{url_host}"

    def __call__(self, *, changed_paths: Sequence[Path]):
        """Generate the documentation using ``sphinx``."""
        if changed_paths:
            cwd = Path.cwd()
            rel_paths = []
            for changed_path in changed_paths[:5]:
                if not changed_path.exists():
                    continue
                with contextlib.suppress(ValueError):
                    changed_path = changed_path.relative_to(cwd)
                rel_paths.append(changed_path.as_posix())
            if rel_paths:
                show_message(f"Detected changes ({', '.join(rel_paths)})")
            show_message("Rebuilding...")

        try:
            for command in self.pre_build_commands:
                show_message("pre-build")
                show_command(command)
                subprocess.run(command, check=True)
        except subprocess.CalledProcessError as e:
            print(f"Pre-build command exited with exit code: {e.returncode}")
            print(
                "Please fix the cause of the error above or press Ctrl+C to stop the "
                "server."
            )
            print(
                "The server will continue serving the build folder, but the contents "
                "being served are no longer in sync with the documentation sources. "
                "Please fix the cause of the error above or press Ctrl+C to stop the "
                "server."
            )
            traceback.print_exception(e)
            return

        if sphinx.version_info[:3] >= (7, 2, 3):
            sphinx_build_args = ["-m", "sphinx", "build"] + self.sphinx_args
        else:
            sphinx_build_args = ["-m", "sphinx"] + self.sphinx_args
        show_command(["python"] + sphinx_build_args)
        try:
            subprocess.run([sys.executable] + sphinx_build_args, check=True)
        except subprocess.CalledProcessError as e:
            print(f"Sphinx exited with exit code: {e.returncode}")
            print(
                "The server will continue serving the build folder, but the contents "
                "being served are no longer in sync with the documentation sources. "
                "Please fix the cause of the error above or press Ctrl+C to stop the "
                "server."
            )
        # Remind the user of the server URL for convenience.
        show_message(f"Serving on {self.uri}")



from __future__ import annotations

import asyncio
from concurrent.futures import ProcessPoolExecutor
from contextlib import AbstractAsyncContextManager, asynccontextmanager
from pathlib import Path
from typing import TYPE_CHECKING

import watchfiles
from starlette.websockets import WebSocket

if TYPE_CHECKING:
    import os
    from collections.abc import Callable, Sequence

    from starlette.types import Receive, Scope, Send

    from sphinx_autobuild.filter import IgnoreFilter


class RebuildServer:
    def __init__(
        self,
        paths: list[os.PathLike[str]],
        ignore_filter: IgnoreFilter,
        change_callback: Callable[[Sequence[Path]], None],
    ) -> None:
        self.paths = [Path(path).resolve(strict=True) for path in paths]
        self.ignore = ignore_filter
        self.change_callback = change_callback
        self.flag = asyncio.Event()
        self.should_exit = asyncio.Event()

    @asynccontextmanager
    async def lifespan(self, _app) -> AbstractAsyncContextManager[None]:
        task = asyncio.create_task(self.main())
        yield
        self.should_exit.set()
        await task
        return

    async def main(self) -> None:
        tasks = (
            asyncio.create_task(self.watch()),
            asyncio.create_task(self.should_exit.wait()),
        )
        done, pending = await asyncio.wait(tasks, return_when=asyncio.FIRST_COMPLETED)
        [task.cancel() for task in pending]
        [task.result() for task in done]

    async def watch(self) -> None:
        async for changes in watchfiles.awatch(
            *self.paths,
            watch_filter=lambda _, path: not self.ignore(path),
        ):
            changed_paths = [Path(path).resolve() for (_, path) in changes]
            with ProcessPoolExecutor() as pool:
                fut = pool.submit(self.change_callback, changed_paths=changed_paths)
                await asyncio.wrap_future(fut)
            self.flag.set()

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        assert scope["type"] == "websocket"
        ws = WebSocket(scope, receive, send)
        await ws.accept()

        tasks = (
            asyncio.create_task(self.watch_reloads(ws)),
            asyncio.create_task(self.wait_client_disconnect(ws)),
        )
        done, pending = await asyncio.wait(tasks, return_when=asyncio.FIRST_COMPLETED)
        [task.cancel() for task in pending]
        [task.result() for task in done]

    async def watch_reloads(self, ws: WebSocket) -> None:
        while True:
            await self.flag.wait()
            self.flag.clear()
            await ws.send_text("refresh")

    @staticmethod
    async def wait_client_disconnect(ws: WebSocket) -> None:
        async for _ in ws.iter_text():
            pass



from sphinx_autobuild.filter import IgnoreFilter


def test_empty():
    ignored = IgnoreFilter([], [])

    assert not ignored("amazing-file.txt")
    assert not ignored("module.pyc")
    assert not ignored("one.rst")
    assert not ignored("two.rst")
    assert not ignored("one.md")
    assert not ignored("foo/random.txt")
    assert not ignored("bar/__pycache__/file.pyc")


def test_single_regex():
    ignored = IgnoreFilter([], [r"\.pyc$"])

    assert not ignored("amazing-file.txt")
    assert ignored("module.pyc")
    assert not ignored("one.rst")
    assert not ignored("two.rst")
    assert not ignored("one.md")
    assert not ignored("foo/random.txt")
    assert ignored("bar/__pycache__/file.pyc")


def test_multiple_regex():
    ignored = IgnoreFilter([], [r"\.md", r"one\.rst"])

    assert not ignored("amazing-file.txt")
    assert not ignored("module.pyc")
    assert ignored("one.rst")
    assert not ignored("two.rst")
    assert ignored("one.md")
    assert not ignored("foo/random.txt")
    assert not ignored("bar/__pycache__/file.pyc")


def test_single_regular():
    ignored = IgnoreFilter(["*.pyc"], [])

    assert not ignored("amazing-file.txt")
    assert ignored("module.pyc")
    assert not ignored("one.rst")
    assert not ignored("two.rst")
    assert not ignored("one.md")
    assert not ignored("foo/random.txt")
    assert ignored("bar/__pycache__/file.pyc")


def test_multiple_regular():
    ignored = IgnoreFilter(["bar", "foo"], [])

    assert not ignored("amazing-file.txt")
    assert not ignored("module.pyc")
    assert not ignored("one.md")
    assert not ignored("one.rst")
    assert not ignored("two.rst")
    assert ignored("foo/random.txt")
    assert ignored("bar/__pycache__/file.pyc")


def test_multiple_both():
    ignored = IgnoreFilter(["bar", "foo"], [r"\.txt", r"one\.*"])

    assert ignored("amazing-file.txt")
    assert not ignored("module.pyc")
    assert ignored("one.md")
    assert ignored("one.rst")
    assert not ignored("two.rst")
    assert ignored("foo/random.txt")
    assert ignored("foo/module.pyc")
    assert ignored("bar/__pycache__/file.pyc")



"""Logic for ignoring paths."""

from __future__ import annotations

import fnmatch
import re
from pathlib import Path


class IgnoreFilter:
    def __init__(self, regular, regex_based):
        """Prepare the function that determines whether a path should be ignored."""
        normalised_paths = [Path(p).resolve().as_posix() for p in regular]
        self.regular_patterns = list(dict.fromkeys(normalised_paths))
        self.regex_based_patterns = [*map(re.compile, dict.fromkeys(regex_based))]

    def __repr__(self):
        return (
            f"IgnoreFilter(regular={self.regular_patterns!r}, "
            f"regex_based={self.regex_based_patterns!r})"
        )

    def __call__(self, filename: str, /):
        """Determine if 'path' should be ignored."""
        normalised_path = Path(filename).resolve().as_posix()
        # Any regular pattern matches.
        for pattern in self.regular_patterns:
            # separators are normalised before creating the IgnoreFilter
            if normalised_path.startswith(f"{pattern}/"):
                return True
            if fnmatch.fnmatch(normalised_path, pattern):
                return True

        # Any regular expression matches.
        for regex in self.regex_based_patterns:  # NoQA: SIM110
            if regex.search(normalised_path):
                return True

        return False



"""Entrypoint for ``python -m sphinx_autobuild``."""

from __future__ import annotations

import argparse
import shlex
import sys
from pathlib import Path

import colorama
import uvicorn

# This isn't public API, but there aren't many better options
from sphinx.cmd.build import get_parser as sphinx_get_parser
from starlette.applications import Starlette
from starlette.middleware import Middleware
from starlette.routing import Mount, WebSocketRoute
from starlette.staticfiles import StaticFiles

from sphinx_autobuild import __version__
from sphinx_autobuild.build import Builder
from sphinx_autobuild.filter import IgnoreFilter
from sphinx_autobuild.middleware import JavascriptInjectorMiddleware
from sphinx_autobuild.server import RebuildServer
from sphinx_autobuild.utils import find_free_port, open_browser, show_message


def main(argv=()):
    """Actual application logic."""
    colorama.just_fix_windows_console()

    if not argv:
        # entry point functions don't receive args
        argv = sys.argv[1:]

    args, build_args = _parse_args(list(argv))

    src_dir = Path(args.sourcedir)
    out_dir = Path(args.outdir)
    out_dir.mkdir(parents=True, exist_ok=True)

    serve_dir = out_dir
    if args.make_mode_builder:
        serve_dir = out_dir / args.make_mode_builder
        serve_dir.mkdir(parents=True, exist_ok=True)

    host_name = args.host
    port_num = args.port or find_free_port()
    url_host = f"{host_name}:{port_num}"

    pre_build_commands = list(map(shlex.split, args.pre_build))

    builder = Builder(
        build_args,
        url_host=url_host,
        pre_build_commands=pre_build_commands,
    )

    watch_dirs = [src_dir] + args.additional_watched_dirs
    ignore_dirs = [
        ".git",
        ".hg",
        ".idea",
        ".mypy_cache",
        "node_modules",
        ".nox",
        ".ruff_cache",
        ".pytest_cache",
        ".pytype",
        ".svn",
        ".tox",
        ".venv",
        "venv",
        ".vscode",
        *args.ignore,
        out_dir,
        args.warnings_file,
        args.doctree_dir,
    ]
    ignore_dirs = list(filter(None, ignore_dirs))
    ignore_handler = IgnoreFilter(ignore_dirs, args.re_ignore)
    app = _create_app(watch_dirs, ignore_handler, builder, serve_dir, url_host)

    if not args.no_initial_build:
        show_message("Starting initial build")
        builder(changed_paths=())

    if args.open_browser:
        open_browser(url_host, args.delay)

    show_message("Waiting to detect changes...")
    try:
        uvicorn.run(app, host=host_name, port=port_num, log_level="warning")
    except KeyboardInterrupt:
        show_message("Server ceasing operations. Cheerio!")


def _create_app(watch_dirs, ignore_handler, builder, out_dir, url_host):
    watcher = RebuildServer(watch_dirs, ignore_handler, change_callback=builder)

    return Starlette(
        routes=[
            WebSocketRoute("/websocket-reload", watcher, name="reload"),
            Mount("/", app=StaticFiles(directory=out_dir, html=True), name="static"),
        ],
        middleware=[Middleware(JavascriptInjectorMiddleware, ws_url=url_host)],
        lifespan=watcher.lifespan,
    )


def _parse_args(argv):
    # Parse once with the Sphinx parser to emit errors
    # and capture the ``-d`` and ``-w`` options.
    # NOTE:
    # The Sphinx parser is not considered to be public API,
    # but as this is a first-party project, we can cheat a little bit.
    sphinx_args = _get_sphinx_build_parser().parse_args(argv.copy())

    # Parse a second time with just our parser
    parser = _get_parser()
    args, build_args = parser.parse_known_args(argv.copy())

    # Copy needed settings
    args.sourcedir = Path(sphinx_args.sourcedir).resolve(strict=True)
    args.outdir = Path(sphinx_args.outputdir).resolve()
    if sphinx_args.doctreedir:
        args.doctree_dir = Path(sphinx_args.doctreedir).resolve()
    else:
        args.doctree_dir = None
    if sphinx_args.warnfile:
        args.warnings_file = Path(sphinx_args.warnfile).resolve()
    else:
        args.warnings_file = None

    # Copy the make-mode builder, if present
    args.make_mode_builder = sphinx_args.use_make_mode or ""

    return args, build_args


def _get_sphinx_build_parser():
    # NOTE:
    # sphinx.cmd.build.get_parser is not considered to be public API,
    # but as this is a first-party project, we can cheat a little bit.
    sphinx_build_parser = sphinx_get_parser()
    sphinx_build_parser.description = None
    sphinx_build_parser.epilog = None
    sphinx_build_parser.prog = "sphinx-autobuild"
    for action in sphinx_build_parser._actions:
        if hasattr(action, "version"):
            # Fix the version
            action.version = f"%(prog)s {__version__}"
            break
    sphinx_build_parser.add_argument(
        "-M",
        dest="use_make_mode",
        help=argparse.SUPPRESS,
    )
    _add_autobuild_arguments(sphinx_build_parser)

    return sphinx_build_parser


def _get_parser():
    """Get the application's argument parser."""
    parser = argparse.ArgumentParser(allow_abbrev=False)
    parser.add_argument(
        "--version", action="version", version=f"sphinx-autobuild {__version__}"
    )
    _add_autobuild_arguments(parser)

    return parser


def _add_autobuild_arguments(parser):
    group = parser.add_argument_group("autobuild options")
    group.add_argument(
        "--port",
        type=int,
        default=8000,
        help="port to serve documentation on. 0 means find and use a free port",
    )
    group.add_argument(
        "--host",
        type=str,
        default="127.0.0.1",
        help="hostname to serve documentation on",
    )
    group.add_argument(
        "--re-ignore",
        action="append",
        default=[],
        help="regular expression for files to ignore, when watching for changes",
    )
    group.add_argument(
        "--ignore",
        action="append",
        default=[],
        help="glob expression for files to ignore, when watching for changes",
    )
    group.add_argument(
        "--no-initial",
        dest="no_initial_build",
        action="store_true",
        default=False,
        help="skip the initial build",
    )
    group.add_argument(
        "--open-browser",
        action="store_true",
        default=False,
        help="open the browser after building documentation",
    )
    group.add_argument(
        "--delay",
        dest="delay",
        type=float,
        default=5,
        help="how long to wait before opening the browser",
    )
    group.add_argument(
        "--watch",
        action="append",
        metavar="DIR",
        default=[],
        help="additional directories to watch",
        dest="additional_watched_dirs",
    )
    group.add_argument(
        "--pre-build",
        action="append",
        metavar="COMMAND",
        default=[],
        help="additional command(s) to run prior to building the documentation",
    )
    return group


if __name__ == "__main__":
    main(sys.argv[1:])



"""A very basic test that the application works."""

import shutil
from pathlib import Path

from starlette.testclient import TestClient

from sphinx_autobuild.__main__ import _create_app
from sphinx_autobuild.build import Builder
from sphinx_autobuild.filter import IgnoreFilter

ROOT = Path(__file__).parent.parent


def test_application(tmp_path):
    src_dir = tmp_path / "docs"
    out_dir = tmp_path / "build"
    shutil.copytree(ROOT / "docs", tmp_path / "docs")
    out_dir.mkdir(parents=True, exist_ok=True)

    url_host = "127.0.0.1:7777"
    ignore_handler = IgnoreFilter([out_dir], [])
    builder = Builder(
        [str(src_dir), str(out_dir)], url_host=url_host, pre_build_commands=[]
    )
    app = _create_app([src_dir], ignore_handler, builder, out_dir, url_host)
    client = TestClient(app)

    builder(changed_paths=())

    response = client.get("/")
    assert response.status_code == 200



from __future__ import annotations

from typing import TYPE_CHECKING

from starlette.datastructures import MutableHeaders

if TYPE_CHECKING:
    from starlette.types import ASGIApp, Message, Receive, Scope, Send


def web_socket_script(ws_url: str) -> str:
    # language=HTML
    return f"""
<script>
const ws = new WebSocket("ws://{ws_url}/websocket-reload");
ws.onmessage = () => window.location.reload();
</script>
"""


class JavascriptInjectorMiddleware:
    def __init__(self, app: ASGIApp, ws_url: str) -> None:
        self.app = app
        self.script = web_socket_script(ws_url).encode("utf-8")

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        add_script = False
        if scope["type"] != "http":
            await self.app(scope, receive, send)
            return

        async def send_wrapper(message: Message) -> None:
            nonlocal add_script
            if message["type"] == "http.response.start":
                headers = MutableHeaders(scope=message)
                if headers.get("Content-Type", "").startswith("text/html"):
                    add_script = True
                    if "Content-Length" in headers:
                        length = int(headers["Content-Length"]) + len(self.script)
                        headers["Content-Length"] = str(length)
            elif message["type"] == "http.response.body":
                request_complete = not message.get("more_body", False)
                if add_script and request_complete:
                    message["body"] += self.script
            await send(message)

        await self.app(scope, receive, send_wrapper)
        return



"""Development automation."""

import nox


@nox.session(reuse_venv=True)
def lint(session):
    session.install("pre-commit")
    session.run("pre-commit", "run", "--all-files", *session.posargs)


@nox.session(python=["3.9", "3.10", "3.11", "3.12", "3.13"])
def test(session):
    session.install("-e", ".[test]", silent=True)
    session.run("pytest", *session.posargs)


@nox.session(reuse_venv=True)
def docs(session):
    session.install("-e", ".", silent=True)
    session.run("sphinx-build", "-b", "html", "docs/", "build/docs")


@nox.session(name="docs-live", reuse_venv=True)
def docs_live(session):
    session.install("-e", ".", silent=True)
    session.run(
        "sphinx-autobuild", "-b", "html", "docs/", "build/docs", *session.posargs
    )



"""Generic utilities."""

from __future__ import annotations

import shlex
import socket
import threading
import time
import webbrowser

from colorama import Fore, Style


def find_free_port():
    """Find and return a free port number.

    Shout-out to https://stackoverflow.com/a/45690594/1931274!
    """
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        s.bind(("", 0))
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        return s.getsockname()[1]


def open_browser(url_host: str, delay: float) -> None:
    def _opener():
        time.sleep(delay)
        webbrowser.open(f"http://{url_host}")

    t = threading.Thread(target=_opener)
    t.start()
    t.join()


def _log(text, *, colour):
    print(f"{Fore.GREEN}[sphinx-autobuild] {colour}{text}{Style.RESET_ALL}")


def show_message(context: str, /) -> None:
    """Show context, with nice formatting and colours."""
    _log(context, colour=Fore.CYAN)


def show_command(command: list[str] | tuple[str, ...], /) -> None:
    """Show command-to-be-executed, with nice formatting and colours."""
    assert isinstance(command, (list, tuple))
    msg = f"> {shlex.join(command)}"
    _log(msg, colour=Fore.BLUE)
